@@ -23,6 +23,7 @@
 #include "expr.h"
 #include "openvswitch/dynamic-string.h"
 #include "openvswitch/hmap.h"
+#include "openvswitch/ofp-actions.h"
 #include "openvswitch/uuid.h"
 #include "util.h"
 #include "ovn/features.h"
@@ -65,8 +66,9 @@ struct collector_set_ids;
     OVNACT(EXCHANGE,          ovnact_move)            \
     OVNACT(DEC_TTL,           ovnact_null)            \
     OVNACT(CT_NEXT,           ovnact_ct_next)         \
-    OVNACT(CT_COMMIT_V1,      ovnact_ct_commit_v1)    \
+    /* CT_COMMIT_V1 is not supported anymore. */      \
     OVNACT(CT_COMMIT_V2,      ovnact_nest)            \
+    OVNACT(CT_COMMIT_TO_ZONE, ovnact_ct_commit_to_zone) \
     OVNACT(CT_DNAT,           ovnact_ct_nat)          \
     OVNACT(CT_SNAT,           ovnact_ct_nat)          \
     OVNACT(CT_DNAT_IN_CZONE,  ovnact_ct_nat)          \
@@ -75,7 +77,7 @@ struct collector_set_ids;
     OVNACT(CT_LB_MARK,        ovnact_ct_lb)           \
     OVNACT(SELECT,            ovnact_select)          \
     OVNACT(CT_CLEAR,          ovnact_null)            \
-    OVNACT(CT_COMMIT_NAT,     ovnact_ct_nat)          \
+    OVNACT(CT_COMMIT_NAT,     ovnact_ct_commit_to_zone) \
     OVNACT(CLONE,             ovnact_nest)            \
     OVNACT(ARP,               ovnact_nest)            \
     OVNACT(ICMP4,             ovnact_nest)            \
@@ -95,6 +97,8 @@ struct collector_set_ids;
     OVNACT(LOOKUP_ND_IP,      ovnact_lookup_mac_bind_ip) \
     OVNACT(PUT_DHCPV4_OPTS,   ovnact_put_opts)        \
     OVNACT(PUT_DHCPV6_OPTS,   ovnact_put_opts)        \
+    OVNACT(DHCPV4_RELAY_REQ_CHK,  ovnact_dhcp_relay)      \
+    OVNACT(DHCPV4_RELAY_RESP_CHK, ovnact_dhcp_relay)      \
     OVNACT(SET_QUEUE,         ovnact_set_queue)       \
     OVNACT(DNS_LOOKUP,        ovnact_result)          \
     OVNACT(LOG,               ovnact_log)             \
@@ -126,6 +130,12 @@ struct collector_set_ids;
     OVNACT(COMMIT_LB_AFF,     ovnact_commit_lb_aff)   \
     OVNACT(CHK_LB_AFF,        ovnact_result)          \
     OVNACT(SAMPLE,            ovnact_sample)          \
+    OVNACT(MAC_CACHE_USE,     ovnact_null)            \
+    OVNACT(CT_ORIG_NW_DST,    ovnact_result)          \
+    OVNACT(CT_ORIG_IP6_DST,   ovnact_result)          \
+    OVNACT(CT_ORIG_TP_DST,    ovnact_result)          \
+    OVNACT(FLOOD_REMOTE,      ovnact_null)            \
+    OVNACT(CT_STATE_SAVE,     ovnact_result)
 
 /* enum ovnact_type, with a member OVNACT_<ENUM> for each action. */
 enum OVS_PACKED_ENUM ovnact_type {
@@ -236,7 +246,7 @@ struct ovnact_next {
 struct ovnact_load {
     struct ovnact ovnact;
     struct expr_field dst;
-    union expr_constant imm;
+    struct expr_constant imm;
 };
 
 /* OVNACT_MOVE, OVNACT_EXCHANGE. */
@@ -255,16 +265,10 @@ struct ovnact_push_pop {
 /* OVNACT_CT_NEXT. */
 struct ovnact_ct_next {
     struct ovnact ovnact;
+    bool dnat_zone;
     uint8_t ltable;                /* Logical table ID of next table. */
 };
 
-/* OVNACT_CT_COMMIT_V1. */
-struct ovnact_ct_commit_v1 {
-    struct ovnact ovnact;
-    uint32_t ct_mark, ct_mark_mask;
-    ovs_be128 ct_label, ct_label_mask;
-};
-
 /* Type of NAT used for the particular action.
  * UNSPEC translates to applying NAT that works for both directions. */
 enum ovnact_ct_nat_type {
@@ -273,7 +277,7 @@ enum ovnact_ct_nat_type {
     OVNACT_CT_NAT_UNSPEC,
 };
 
-/* OVNACT_CT_DNAT, OVNACT_CT_SNAT, OVNACT_CT_COMMIT_NAT. */
+/* OVNACT_CT_DNAT, OVNACT_CT_SNAT. */
 struct ovnact_ct_nat {
     struct ovnact ovnact;
     int family;
@@ -295,6 +299,15 @@ struct ovnact_ct_nat {
     uint8_t ltable;             /* Logical table ID of next table. */
 };
 
+/* OVNACT_CT_COMMIT_TO_ZONE, OVNACT_CT_COMMIT_NAT. */
+struct ovnact_ct_commit_to_zone {
+    struct ovnact ovnact;
+
+    bool dnat_zone;
+    bool do_nat;
+    uint8_t ltable;
+};
+
 enum ovnact_ct_lb_flag {
     OVNACT_CT_LB_FLAG_NONE,
     OVNACT_CT_LB_FLAG_SKIP_SNAT,
@@ -331,6 +344,7 @@ struct ovnact_select {
     struct ovnact_select_dst *dsts;
     size_t n_dsts;
     uint8_t ltable;             /* Logical table ID of next table. */
+    char *hash_fields;
     struct expr_field res_field;
 };
 
@@ -386,6 +400,14 @@ struct ovnact_put_opts {
     size_t n_options;
 };
 
+/* OVNACT_DHCP_RELAY. */
+struct ovnact_dhcp_relay {
+    struct ovnact ovnact;
+    struct expr_field dst;      /* 1-bit destination field. */
+    ovs_be32 relay_ipv4;
+    ovs_be32 server_ipv4;
+};
+
 /* Valid arguments to SET_QUEUE action.
  *
  * QDISC_MIN_QUEUE_ID is the default queue, so user-defined queues should
@@ -399,10 +421,11 @@ struct ovnact_set_queue {
     uint16_t queue_id;
 };
 
-/* OVNACT_DNS_LOOKUP, OVNACT_CHK_LB_HAIRPIN, OVNACT_CHK_LB_HAIRPIN_REPLY. */
+/* OVNACT_DNS_LOOKUP, OVNACT_CHK_LB_HAIRPIN, OVNACT_CHK_LB_HAIRPIN_REPLY,
+ * OVNACT_CT_ORIG_NW_DST, CT_ORIG_IP6_DST, CT_ORIG_TP_DST */
 struct ovnact_result {
     struct ovnact ovnact;
-    struct expr_field dst;      /* 1-bit destination field. */
+    struct expr_field dst;      /* destination field. */
 };
 
 /* OVNACT_LOG. */
@@ -483,13 +506,15 @@ struct ovnact_lookup_fdb {
 /* OVNACT_SAMPLE */
 struct ovnact_sample {
     struct ovnact ovnact;
-    uint16_t probability;       /* probability over UINT16_MAX. */
-    uint8_t obs_domain_id;      /* most significant byte of the
-                                   observation domain id. The other 24 bits
-                                   will come from the datapath's tunnel key. */
-    uint32_t collector_set_id;  /* colector_set_id. */
-    uint32_t obs_point_id;      /* observation point id. */
-    bool use_cookie;            /* use cookie as obs_point_id */
+    uint16_t probability;               /* probability over UINT16_MAX. */
+    uint8_t obs_domain_id;              /* most significant byte of the
+                                           observation domain id. The other
+                                           24 bits will come from the
+                                           datapath's tunnel key. */
+    uint32_t collector_set_id;          /* colector_set_id. */
+    struct expr_field obs_point_id_src; /* observation point id source reg */
+    uint32_t obs_point_id;              /* observation point id */
+    bool use_cookie;                    /* use cookie as obs_point_id */
 };
 
 /* OVNACT_COMMIT_ECMP_NH. */
@@ -513,6 +538,19 @@ struct ovnact_commit_lb_aff {
     uint16_t timeout;
 };
 
+#define OVN_FIELD_NOTE_MAGIC "ovn"
+
+struct ovn_field_note_header {
+    OFPACT_PADDED_MEMBERS(
+        char magic[4];
+        ovs_be16 type;   /* The type of ovn field note, based
+                          * on 'enum ovn_field_id'. */
+    );
+    uint8_t data[];
+};
+
+BUILD_ASSERT_DECL(sizeof(struct ovn_field_note_header) == 8);
+
 /* Internal use by the helpers below. */
 void ovnact_init(struct ovnact *, enum ovnact_type, size_t len);
 void *ovnact_put(struct ofpbuf *, enum ovnact_type, size_t len);
@@ -571,13 +609,21 @@ void *ovnact_put(struct ofpbuf *, enum ovnact_type, size_t len);
 OVNACTS
 #undef OVNACT
 
-enum action_opcode {
+/* List of OVN logical action opcodes.
+ *
+ * This macro is used directly only internally by this header and actions.c,
+ * but the list is still of interest to developers.
+ *
+ * Each ACTION_OPCODE invocation has the <ENUM> parameter, used below in the
+ * enum definition of ACTION_OPCODE_<ENUM>, and elsewhere.
+ */
+#define ACTION_OPCODES                                                        \
     /* "arp { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_ARP,
-
+     */                                                                       \
+    ACTION_OPCODE(ARP)                                                        \
+                                                                              \
     /* "put_arp(port, ip, mac)"
      *
      * Arguments are passed through the packet metadata and data, as follows:
@@ -585,9 +631,9 @@ enum action_opcode {
      *     MFF_REG0 = ip
      *     MFF_LOG_INPORT = port
      *     MFF_ETH_SRC = mac
-     */
-    ACTION_OPCODE_PUT_ARP,
-
+     */                                                                       \
+    ACTION_OPCODE(PUT_ARP)                                                    \
+                                                                              \
     /* "result = put_dhcp_opts(offer_ip, option, ...)".
      *
      * Arguments follow the action_header, in this format:
@@ -595,15 +641,15 @@ enum action_opcode {
      *   - A 32-bit integer specifying a bit offset within the result field.
      *   - The 32-bit DHCP offer IP.
      *   - Any number of DHCP options.
-     */
-    ACTION_OPCODE_PUT_DHCP_OPTS,
-
+     */                                                                       \
+    ACTION_OPCODE(PUT_DHCP_OPTS)                                              \
+                                                                              \
     /* "nd_na { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_ND_NA,
-
+     */                                                                       \
+    ACTION_OPCODE(ND_NA)                                                      \
+                                                                              \
     /* "put_nd(port, ip6, mac)"
      *
      * Arguments are passed through the packet metadata and data, as follows:
@@ -611,141 +657,165 @@ enum action_opcode {
      *     MFF_XXREG0 = ip6
      *     MFF_LOG_INPORT = port
      *     MFF_ETH_SRC = mac
-     */
-    ACTION_OPCODE_PUT_ND,
-
+     */                                                                       \
+    ACTION_OPCODE(PUT_ND)                                                     \
+                                                                              \
     /* "result = put_dhcpv6_opts(option, ...)".
      *
      * Arguments follow the action_header, in this format:
      *   - A 32-bit or 64-bit OXM header designating the result field.
      *   - A 32-bit integer specifying a bit offset within the result field.
      *   - Any number of DHCPv6 options.
-     */
-    ACTION_OPCODE_PUT_DHCPV6_OPTS,
-
+     */                                                                       \
+    ACTION_OPCODE(PUT_DHCPV6_OPTS)                                            \
+                                                                              \
     /* "result = dns_lookup()".
      * Arguments follow the action_header, in this format:
      *   - A 32-bit or 64-bit OXM header designating the result field.
      *   - A 32-bit integer specifying a bit offset within the result field.
      *
-     */
-    ACTION_OPCODE_DNS_LOOKUP,
-
+     */                                                                       \
+    ACTION_OPCODE(DNS_LOOKUP)                                                 \
+                                                                              \
     /* "log(arguments)".
      *
      * Arguments are as follows:
      *   - An 8-bit verdict.
      *   - An 8-bit severity.
      *   - A variable length string containing the name.
-     */
-    ACTION_OPCODE_LOG,
-
+     */                                                                       \
+    ACTION_OPCODE(LOG)                                                        \
+                                                                              \
     /* "result = put_nd_ra_opts(option, ...)".
      * Arguments follow the action_header, in this format:
      *   - A 32-bit or 64-bit OXM header designating the result field.
      *   - A 32-bit integer specifying a bit offset within the result field.
      *   - Any number of ICMPv6 options.
-     */
-    ACTION_OPCODE_PUT_ND_RA_OPTS,
-
+     */                                                                       \
+    ACTION_OPCODE(PUT_ND_RA_OPTS)                                             \
+                                                                              \
     /* "nd_ns { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_ND_NS,
-
+     */                                                                       \
+    ACTION_OPCODE(ND_NS)                                                      \
+                                                                              \
     /* "icmp4 { ...actions... } and icmp6 { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_ICMP,
-
+     */                                                                       \
+    ACTION_OPCODE(ICMP)                                                       \
+                                                                              \
     /* "tcp_reset { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_TCP_RESET,
-
+     */                                                                       \
+    ACTION_OPCODE(TCP_RESET)                                                  \
+                                                                              \
     /* "nd_na_router { ...actions... }" with rso flag 'ND_RSO_ROUTER' set.
-        *
-        * The actions, in OpenFlow 1.3 format, follow the action_header.
-        */
-    ACTION_OPCODE_ND_NA_ROUTER,
-
-     /* MTU value (to put in the icmp4 header field - frag_mtu) follow the
-     * action header. */
-    ACTION_OPCODE_PUT_ICMP4_FRAG_MTU,
-
+     *
+     * The actions, in OpenFlow 1.3 format, follow the action_header.
+     */                                                                       \
+    ACTION_OPCODE(ND_NA_ROUTER)                                               \
+                                                                              \
+    /* MTU value (to put in the icmp4 header field - frag_mtu) follow the
+     * action header. */                                                      \
+    ACTION_OPCODE(PUT_ICMP4_FRAG_MTU)                                         \
+                                                                              \
     /* "icmp4_error { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_ICMP4_ERROR,
-
-    /* "trigger_event (event_type)" */
-    ACTION_OPCODE_EVENT,
-
+     */                                                                       \
+    ACTION_OPCODE(ICMP4_ERROR)                                                \
+                                                                              \
+    /* "trigger_event (event_type)" */                                        \
+    ACTION_OPCODE(EVENT)                                                      \
     /* "igmp".
      *
      * Snoop IGMP, learn the multicast participants
-     */
-    ACTION_OPCODE_IGMP,
-
+     */                                                                       \
+    ACTION_OPCODE(IGMP)                                                       \
+                                                                              \
     /* "bind_vport(vport, vport_parent)".
      *
      *   'vport' follows the action_header, in the format - 32-bit field.
      *   'vport_parent' is passed through the packet metadata as
      *    MFF_LOG_INPORT.
-     */
-    ACTION_OPCODE_BIND_VPORT,
-
+     */                                                                       \
+    ACTION_OPCODE(BIND_VPORT)                                                 \
+                                                                              \
     /* "handle_svc_check(port)"."
      *
      * Arguments are passed through the packet metadata and data, as follows:
      *
      *     MFF_LOG_INPORT = port
-     */
-    ACTION_OPCODE_HANDLE_SVC_CHECK,
+     */                                                                       \
+    ACTION_OPCODE(HANDLE_SVC_CHECK)                                           \
+                                                                              \
     /* handle_dhcpv6_reply { ...actions ...}."
      *
      *  The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_DHCP6_SERVER,
-
+     */                                                                       \
+    ACTION_OPCODE(DHCP6_SERVER)                                               \
+                                                                              \
     /* "icmp6_error { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_ICMP6_ERROR,
-
+     */                                                                       \
+    ACTION_OPCODE(ICMP6_ERROR)                                                \
+                                                                              \
     /* MTU value (to put in the icmp6 header field - frag_mtu) follow the
-     * action header. */
-    ACTION_OPCODE_PUT_ICMP6_FRAG_MTU,
-
+     * action header. */                                                      \
+    ACTION_OPCODE(PUT_ICMP6_FRAG_MTU)                                         \
+                                                                              \
     /* "reject { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_REJECT,
-
+     */                                                                       \
+    ACTION_OPCODE(REJECT)                                                     \
+                                                                              \
     /* handle_bfd_msg { ...actions ...}."
      *
      *  The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_BFD_MSG,
-
+     */                                                                       \
+    ACTION_OPCODE(BFD_MSG)                                                    \
+                                                                              \
     /* "sctp_abort { ...actions... }".
      *
      * The actions, in OpenFlow 1.3 format, follow the action_header.
-     */
-    ACTION_OPCODE_SCTP_ABORT,
+     */                                                                       \
+    ACTION_OPCODE(SCTP_ABORT)                                                 \
+                                                                              \
+    /* put_fdb(inport, eth.src). */                                           \
+    ACTION_OPCODE(PUT_FDB)                                                    \
+                                                                              \
+    /* activation_strategy_rarp() */                                          \
+    ACTION_OPCODE(ACTIVATION_STRATEGY_RARP)                                   \
+                                                                              \
+    /*  DEPRECATED: split buffer action. */                                   \
+    ACTION_OPCODE(SPLIT_BUF_ACTION)                                           \
+                                                                              \
+    /* "dhcp_relay_req_chk(relay_ip, server_ip)".
+     *
+     * Arguments follow the action_header, in this format:
+     *   - The 32-bit DHCP relay IP.
+     *   - The 32-bit DHCP server IP.
+     */                                                                       \
+    ACTION_OPCODE(DHCP_RELAY_REQ_CHK)                                         \
+                                                                              \
+    /* "dhcp_relay_resp_chk(relay_ip, server_ip)".
+     *
+     * Arguments follow the action_header, in this format:
+     *   - The 32-bit DHCP relay IP.
+     *   - The 32-bit DHCP server IP.
+     */                                                                       \
+    ACTION_OPCODE(DHCP_RELAY_RESP_CHK)
 
-    /* put_fdb(inport, eth.src).
-     */
-    ACTION_OPCODE_PUT_FDB,
 
-    /* activation_strategy_rarp() */
-    ACTION_OPCODE_ACTIVATION_STRATEGY_RARP,
+enum action_opcode {
+#define ACTION_OPCODE(ENUM) ACTION_OPCODE_##ENUM,
+    ACTION_OPCODES
+#undef ACTION_OPCODE
 };
 
 /* Header. */
@@ -755,12 +825,6 @@ struct action_header {
 };
 BUILD_ASSERT_DECL(sizeof(struct action_header) == 8);
 
-OVS_PACKED(
-struct ovnfield_act_header {
-    ovs_be16 id; /* one of enum ovnfield_id. */
-    ovs_be16 len; /* Length of the ovnfield data. */
-});
-
 struct ovnact_parse_params {
     /* A table of "struct expr_symbol"s to support (as one would provide to
      * expr_parse()). */
@@ -841,6 +905,12 @@ struct ovnact_encode_params {
     /* The datapath key. */
     uint32_t dp_key;
 
+    /* Indication if we should add explicit output after arp/nd_ns action. */
+    bool explicit_arp_ns_output;
+
+    /* Indication if we should use the "old" version of registers. */
+    bool register_consolidation;
+
     /* OVN maps each logical flow table (ltable), one-to-one, onto a physical
      * OpenFlow flow table (ptable).  A number of parameters describe this
      * mapping and data related to flow tables:
@@ -885,6 +955,18 @@ struct ovnact_encode_params {
                                    sends packets to controller. */
     uint32_t common_nat_ct_zone; /* When performing NAT in a common CT zone,
                                     this determines which CT zone to use */
+    uint32_t mac_cache_use_table; /* OpenFlow table for 'mac_cache_use'
+                                   * to resubmit. */
+    uint32_t ct_nw_dst_load_table; /* OpenFlow table for 'ct_nw_dst'
+                                    *  to resubmit. */
+    uint32_t ct_ip6_dst_load_table; /* OpenFlow table for 'ct_ip6_dst'
+                                     *  to resubmit. */
+    uint32_t ct_tp_dst_load_table; /* OpenFlow table for 'ct_tp_dst'
+                                    *  to resubmit. */
+    uint32_t flood_remote_table; /* OpenFlow table for 'chassis_flood'
+                                  * to resubmit. */
+    uint32_t ct_state_save_table; /* OpenFlow table for saving ct_state to
+                                   * register. */
 };
 
 void ovnacts_encode(const struct ovnact[], size_t ovnacts_len,
@@ -895,7 +977,13 @@ void ovnacts_free(struct ovnact[], size_t ovnacts_len);
 char *ovnact_op_to_string(uint32_t);
 int encode_ra_dnssl_opt(char *data, char *buf, int buf_len);
 
+size_t encode_start_controller_op(enum action_opcode opcode, bool pause,
+                                  uint32_t meter_id, struct ofpbuf *ofpacts);
+void encode_finish_controller_op(size_t ofs, struct ofpbuf *ofpacts);
 void encode_controller_op(enum action_opcode opcode, uint32_t meter_id,
                           struct ofpbuf *ofpacts);
 
+size_t encode_start_ovn_field_note(enum ovn_field_id, struct ofpbuf *ofpacts);
+void encode_finish_ovn_field_note(size_t offset, struct ofpbuf *ofpacts);
+
 #endif /* ovn/actions.h */
