@@ -127,9 +127,9 @@ static void
 advertised_route_table_sync(
     struct ovsdb_idl_txn *ovnsb_txn,
     const struct sbrec_advertised_route_table *sbrec_advertised_route_table,
-    const struct lr_stateful_table *lr_stateful_table,
     const struct hmap *routes,
     const struct hmap *dynamic_routes,
+    const struct hmap *ls_ports,
     struct advertised_route_sync_data *data);
 
 bool
@@ -242,20 +242,18 @@ en_advertised_route_sync_run(struct engine_node *node, void *data OVS_UNUSED)
         = engine_get_input_data("routes", node);
     struct dynamic_routes_data *dynamic_routes_data
         = engine_get_input_data("dynamic_routes", node);
+    struct northd_data *northd_data = engine_get_input_data("northd", node);
     const struct engine_context *eng_ctx = engine_get_context();
     const struct sbrec_advertised_route_table *sbrec_advertised_route_table =
         EN_OVSDB_GET(engine_get_input("SB_advertised_route", node));
-    struct ed_type_lr_stateful *lr_stateful_data =
-        engine_get_input_data("lr_stateful", node);
 
     stopwatch_start(ADVERTISED_ROUTE_SYNC_RUN_STOPWATCH_NAME, time_msec());
 
     advertised_route_table_sync(eng_ctx->ovnsb_idl_txn,
                                 sbrec_advertised_route_table,
-                                &lr_stateful_data->table,
                                 &routes_data->parsed_routes,
                                 &dynamic_routes_data->routes,
-                                routes_sync_data);
+                                &northd_data->ls_ports, routes_sync_data);
 
     stopwatch_stop(ADVERTISED_ROUTE_SYNC_RUN_STOPWATCH_NAME, time_msec());
     engine_set_node_state(node, EN_UPDATED);
@@ -569,40 +567,37 @@ publish_lport_addresses(struct hmap *sync_routes,
     }
 }
 
-/* Collect all IP addresses connected via this LRP. */
 static void
-publish_host_routes_lrp(struct hmap *sync_routes,
-                        const struct lr_stateful_table *lr_stateful_table,
-                        const struct ovn_datapath *advertising_od,
-                        const struct ovn_port *advertising_op,
-                        struct advertised_route_sync_data *data,
-                        const struct ovn_port *lrp)
+publish_host_routes_for_virtual_ports(
+    struct ovn_port *port,
+    const struct hmap *ls_ports,
+    const struct ovn_datapath *advertising_od,
+    const struct ovn_port *advertising_op,
+    struct hmap *sync_routes)
 {
-    /* This is a LSP connected to an LRP */
-    const struct lport_addresses *addresses = &lrp->lrp_networks;
-    publish_lport_addresses(sync_routes, advertising_od,
-                            advertising_op, addresses, lrp);
+    ovs_assert(port->sb);
 
-    const struct lr_stateful_record *lr_stateful_rec;
-    lr_stateful_rec =
-        lr_stateful_table_find_by_index(lr_stateful_table, lrp->od->index);
-    /* We also need to track this LR as we need to recompute when
-     * any of its IPs change. */
-    uuidset_insert(&data->nb_lr, &lr_stateful_rec->nbr_uuid);
-    struct ovn_port_routable_addresses addrs =
-        get_op_addresses(lrp, lr_stateful_rec, false);
-    for (size_t i = 0; i < addrs.n_addrs; i++) {
-        publish_lport_addresses(sync_routes, advertising_od,
-                                advertising_op, &addrs.laddrs[i], lrp);
+    const char *virtual_parent = port->sb->virtual_parent;
+    if (!virtual_parent) {
+        return;
+    }
+
+    struct ovn_port *vp = ovn_port_find(ls_ports, virtual_parent);
+    if (!vp) {
+        return;
+    }
+
+    for (size_t i = 0; i < port->n_lsp_addrs; i++) {
+        publish_lport_addresses(sync_routes, advertising_od, advertising_op,
+                                &port->lsp_addrs[i], vp);
     }
-    destroy_routable_addresses(&addrs);
 }
 
 /* Collect all IP addresses connected to the out_port of a route.
  * This traverses all LSPs on the LS connected to the out_port. */
 static void
 publish_host_routes(struct hmap *sync_routes,
-                    const struct lr_stateful_table *lr_stateful_table,
+                    const struct hmap *ls_ports,
                     const struct ovn_datapath *advertising_od,
                     const struct ovn_port *advertising_op,
                     struct advertised_route_sync_data *data)
@@ -618,8 +613,9 @@ publish_host_routes(struct hmap *sync_routes,
 
     if (peer_od->nbr) {
         /* This is a LRP directly connected to another LRP. */
-        publish_host_routes_lrp(sync_routes, lr_stateful_table, advertising_od,
-                                advertising_op, data, advertising_op->peer);
+        const struct ovn_port *lrp = advertising_op->peer;
+        publish_lport_addresses(sync_routes, advertising_od,
+                                advertising_op, &lrp->lrp_networks, lrp);
         return;
     }
 
@@ -631,9 +627,13 @@ publish_host_routes(struct hmap *sync_routes,
     HMAP_FOR_EACH (port, dp_node, &peer_od->ports) {
         if (port->peer && port->peer->nbrp) {
             /* This is a LSP connected to an LRP */
-            publish_host_routes_lrp(sync_routes, lr_stateful_table,
-                                    advertising_od, advertising_op,
-                                    data, port->peer);
+            const struct ovn_port *lrp = port->peer;
+            publish_lport_addresses(sync_routes, advertising_od,
+                                    advertising_op, &lrp->lrp_networks, lrp);
+        } else if (port->nbsp && !strcmp(port->nbsp->type, "virtual")) {
+            publish_host_routes_for_virtual_ports(port, ls_ports,
+                                                  advertising_od,
+                                                  advertising_op, sync_routes);
         } else {
             /* This is just a plain LSP */
             for (size_t i = 0; i < port->n_lsp_addrs; i++) {
@@ -686,38 +686,46 @@ should_advertise_route(const struct uuidset *host_route_lrps,
     }
 }
 
-/* Returns true if the route has already been fully processed for
- * advertising (e.g., for connected routes when connected-as-host is set).
- * Returns false otherwise, in which case the a new 'ar_entry' needs to
- * be created for this route.
- */
+/* Returns true if the connected route was advertised as a set of host routes
+ * (/32 for IPv4 and /128 for IPv6), one for each individual IP known to be
+ * reachable in the connected route's subnet.  Returns false otherwise. */
 static bool
-process_prereqs_advertise_route(
-    const struct lr_stateful_table *lr_stateful_table,
+advertise_routes_as_host_prefix(
     struct advertised_route_sync_data *data,
     struct uuidset *host_route_lrps,
     struct hmap *sync_routes,
+    const struct hmap *ls_ports,
     const struct ovn_datapath *advertising_od,
     const struct ovn_port *advertising_op,
-    const struct ovn_port *tracked_op,
-    enum route_source source)
+    enum route_source source
+)
 {
+    if (source != ROUTE_SOURCE_CONNECTED) {
+        return false;
+    }
+
     enum dynamic_routing_redistribute_mode drr =
         advertising_op->dynamic_routing_redistribute;
+    if (!drr_mode_CONNECTED_AS_HOST_is_set(drr)) {
+        return false;
+    }
 
+    uuidset_insert(host_route_lrps, &advertising_op->nbrp->header_.uuid);
+    publish_host_routes(sync_routes, ls_ports, advertising_od,
+                        advertising_op, data);
+    return true;
+}
+
+/* Track datapaths (routers/switches) whose changes should trigger
+ * the set of advertised routes.  That includes NAT and LB related
+ * advertised routes. */
+static void
+advertise_route_track_od(struct advertised_route_sync_data *data,
+                         const struct ovn_datapath *advertising_od,
+                         const struct ovn_port *tracked_op,
+                         enum route_source source)
+{
     switch (source) {
-    case ROUTE_SOURCE_CONNECTED:
-        if (drr_mode_CONNECTED_AS_HOST_is_set(drr)) {
-            const struct uuid *lrp_uuid = &advertising_op->nbrp->header_.uuid;
-            uuidset_insert(host_route_lrps, lrp_uuid);
-            publish_host_routes(sync_routes, lr_stateful_table,
-                                advertising_od, advertising_op,
-                                data);
-            return true;
-        }
-        break;
-    case ROUTE_SOURCE_STATIC:
-        break;
     case ROUTE_SOURCE_NAT:
         /* If NAT route tracks port on a different DP than the one that
          * advertises the route, we need to watch for changes on that DP as
@@ -741,21 +749,23 @@ process_prereqs_advertise_route(
                            &tracked_op->od->nbr->header_.uuid);
         }
         break;
+    case ROUTE_SOURCE_CONNECTED:
+    case ROUTE_SOURCE_STATIC:
+        break;
     case ROUTE_SOURCE_LEARNED:
         OVS_NOT_REACHED();
     default:
         OVS_NOT_REACHED();
     }
-    return false;
 }
 
 static void
 advertised_route_table_sync(
     struct ovsdb_idl_txn *ovnsb_txn,
     const struct sbrec_advertised_route_table *sbrec_advertised_route_table,
-    const struct lr_stateful_table *lr_stateful_table,
     const struct hmap *routes,
     const struct hmap *dynamic_routes,
+    const struct hmap *ls_ports,
     struct advertised_route_sync_data *data)
 {
     struct hmap sync_routes = HMAP_INITIALIZER(&sync_routes);
@@ -768,25 +778,36 @@ advertised_route_table_sync(
             continue;
         }
 
-        if (prefix_is_link_local(&route->prefix, route->plen)) {
-            continue;
-        }
-
         if (!should_advertise_route(&host_route_lrps, route->od,
                                     route->out_port, route->source)) {
             continue;
         }
 
-        if (process_prereqs_advertise_route(lr_stateful_table, data,
-                                            &host_route_lrps, &sync_routes,
+        if (advertise_routes_as_host_prefix(data, &host_route_lrps,
+                                            &sync_routes, ls_ports,
                                             route->od, route->out_port,
-                                            route->tracked_port,
                                             route->source)) {
             continue;
         }
 
+        if (prefix_is_link_local(&route->prefix, route->plen)) {
+            continue;
+        }
+
+        advertise_route_track_od(data, route->od, route->tracked_port,
+                                 route->source);
+
+        const struct sbrec_port_binding *tracked_port =
+            route->tracked_port ? route->tracked_port->sb : NULL;
+        char *ip_prefix = normalize_v46_prefix(&route->prefix, route->plen);
+        if (ar_entry_find(&sync_routes, route->od->sb,
+                          route->out_port->sb, ip_prefix,
+                          tracked_port)) {
+            free(ip_prefix);
+            continue;
+        }
         ar_entry_add_nocopy(&sync_routes, route->od, route->out_port,
-                            normalize_v46_prefix(&route->prefix, route->plen),
+                            ip_prefix,
                             route->tracked_port,
                             route->source);
     }
@@ -799,14 +820,18 @@ advertised_route_table_sync(
             continue;
         }
 
-        if (process_prereqs_advertise_route(lr_stateful_table, data,
-                                            &host_route_lrps, &sync_routes,
-                                            route_e->od, route_e->op,
-                                            route_e->tracked_port,
-                                            route_e->source)) {
+        advertise_route_track_od(data, route_e->od, route_e->tracked_port,
+                                 route_e->source);
+
+        const struct sbrec_port_binding *tracked_pb =
+            route_e->tracked_port ? route_e->tracked_port->sb : NULL;
+        if (ar_entry_find(&sync_routes, route_e->od->sb, route_e->op->sb,
+                          route_e->ip_prefix, tracked_pb)) {
+            /* We could already have advertised route entry for LRP IP that
+             * corresponds to "snat" when "connected-as-host" is combined
+             * with "nat". Skip it. */
             continue;
         }
-
         ar_entry_add(&sync_routes, route_e->od, route_e->op,
                      route_e->ip_prefix, route_e->tracked_port,
                      route_e->source);
@@ -819,12 +844,17 @@ advertised_route_table_sync(
         route_e = ar_entry_find(&sync_routes, sb_route->datapath,
                                 sb_route->logical_port, sb_route->ip_prefix,
                                 sb_route->tracked_port);
-        if (route_e) {
-          hmap_remove(&sync_routes, &route_e->hmap_node);
-          ar_entry_free(route_e);
-        } else {
-          sbrec_advertised_route_delete(sb_route);
+        if (!route_e) {
+            sbrec_advertised_route_delete(sb_route);
+            continue;
         }
+
+        if (route_e->tracked_port && !sb_route->tracked_port) {
+            sbrec_advertised_route_set_tracked_port(
+                sb_route, route_e->tracked_port->sb);
+        }
+        hmap_remove(&sync_routes, &route_e->hmap_node);
+        ar_entry_free(route_e);
     }
 
     HMAP_FOR_EACH_POP (route_e, hmap_node, &sync_routes) {
